/*
 * PWM Controller.cpp
 * Author : A_C_E
 */ 

#include <avr/io.h>
#include <avr/interrupt.h>

//Pin defines Master/Slave
#define PIND_MASTER PD6 //in
#define PIND_OPEN PD5 //in
#define PIND_SENSE_SPEED_II PD4 //in
#define PIND_SENSE_SPEED_I PD3 //in

//Pin defines Master
#define PINB_NC_1 PB4 //in, no function
#define PINB_NC_2 PB3 //in, no function
#define PINB_NC_3 PB2 //in, no function
#define PINB_NC_4 PB1 //in, no function
#define PINB_SUMMER PB0 //out
#define PIND_V_OUT PD2 //out

//Pin defines Slave
#define PINB_PWM2 PB4 //out
#define PINB_PWM1 PB3 //out
#define PINB_FAN2ON PB2 //out
#define PINB_FAN1ON PB1 //out
#define PINB_ERR PB0 //out
#define PIND_V_IN PD2 //in

#define revCnt 7 //==7 * 10s
#define revCntSummer 90 //==90 * 10s

#define connectTimeout 3;

//time system fields
volatile uint8_t systemTimer; //master = 10s, slave = 1s
volatile uint8_t ms_100;
volatile uint8_t flag_ms_1;

//flags I/O master/slave
volatile uint8_t in_master;
volatile uint8_t in_Sense_Speed_I;
volatile uint8_t in_Sense_Speed_II;

//flags I/O master
volatile uint8_t in_Sense_Speed_I_shift;

//flags I/O slave
volatile uint8_t in_V_IN;

//flags master/slave
volatile uint8_t chkPeriode;
volatile uint8_t flag_in_Sense_Speed;

//flags master
volatile uint8_t modeSummerCnt;
volatile uint8_t flag_modeSummer;
volatile uint8_t reverseCnt;
volatile uint8_t flag_reverse;
volatile uint8_t modeSummerTestGate;

//flags slave
volatile uint8_t flag_in_V_IN;
volatile uint8_t dutyCycles[8]={255,155,120,0,120,155,255,0}; //Hi_Bwd, Mid_Bwd, Lo_Bwd, Off_Std, Lo_Fwd, Mid_Fwd, Hi_Fwd, Off_Err (<20 || >230)
volatile uint8_t targetSpeed;
volatile uint8_t dutyCycle1;
volatile uint8_t dutyCycle2;
volatile uint8_t ramp;
volatile uint8_t connected;
volatile uint8_t flag_in_Sense_Speed_I;
volatile uint8_t flag_in_Sense_Speed_II;
volatile uint8_t senseChanged_I;
volatile uint8_t senseChanged_II;
volatile uint8_t fanBlocked_I;
volatile uint8_t fanBlocked_II;

void masterMain(void)
{
    if(!flag_in_Sense_Speed&&in_Sense_Speed_I) //check for summer mode
    {
        if(modeSummerTestGate) {modeSummerCnt++;}
        if(modeSummerCnt>3)
        {
            modeSummerCnt=0;
            flag_modeSummer=(unsigned)~flag_modeSummer;
            if(flag_modeSummer) {PORTB|=(1<<PINB_SUMMER);}
            else {PORTB&=(unsigned)~(1<<PINB_SUMMER);}
        }
        modeSummerTestGate=5;
    }
    flag_in_Sense_Speed=in_Sense_Speed_I;

    int8_t tsp=(in_Sense_Speed_I|(in_Sense_Speed_II<<1));
    if(!flag_reverse) {tsp=-tsp;}             
    if (!systemTimer)
    {
        systemTimer=100;
        if(tsp)
        {
            if(reverseCnt) {reverseCnt--;}
            else
            {
                flag_reverse=(unsigned)~flag_reverse;
                reverseCnt=revCnt;
                if(!flag_reverse) {if(flag_modeSummer) {reverseCnt=revCntSummer;}}
            }
        }
    }    
    uint8_t pulseTo=125+(30*tsp);
    // fwd high = 205.. 225 ms, 86%
    // fwd mid = 175..195 ms, 74%
    // fwd slow = 145..165 ms, 62%
    // stop = 125ms, 50%
    // bwd slow = 85..105 ms, 38%
    // bwd mid = 55..75 ms, 26%
    // bwd high = 25..45 ms, 14%
    if(chkPeriode>=pulseTo) {PORTD|=(1<<PIND_V_OUT);}
    else {PORTD&=(unsigned)~(1<<PIND_V_OUT);}
}

void slaveMain(void)
{
    uint8_t ct=connected;
    if(!systemTimer) {if(ct) {ct--;}}
    connected=ct;

    if(!in_V_IN)
    {
        if(flag_in_V_IN)
        {
            connected=connectTimeout;
            uint8_t cp=chkPeriode;
            uint8_t tsp=7;
            if(cp>20) {tsp=((cp-20)/30);}
            targetSpeed=tsp;
        }        
        chkPeriode=0;
    }    
    flag_in_V_IN=in_V_IN;
    ct=connected;
    if(!ct) {targetSpeed=7;}
    if(!ct||fanBlocked_I||fanBlocked_II) {PORTB|=(1<<PINB_ERR);}
    else {PORTB&=(unsigned)~(1<<PINB_ERR);}

    uint8_t tDC1=0;
    uint8_t tDC2=0;
    uint8_t tsp=targetSpeed;
    if(tsp>3) {tDC1=dutyCycles[tsp];}
    else {tDC2=dutyCycles[tsp];}

    uint8_t dC1=dutyCycle1;
    uint8_t dC2=dutyCycle2;
    if(ramp++>40)
    {
        ramp=0;
        if(dC1!=tDC1)
        {
            if(dC1<tDC1) {dC1=tDC1;} else {dC1--;}
            dutyCycle1=dC1;
        }
        if(dC2!=tDC2)
        {
            if(dC2<tDC2) {dC2=tDC2;} else {dC2--;}
            dutyCycle2=dC2;
        }
    }
   
    if(dC1) {PORTB|=(1<<PINB_FAN1ON);}
    else {PORTB&=(unsigned)~(1<<PINB_FAN1ON);}
    OCR1A=dC1;

    if(dC2) {PORTB|=(1<<PINB_FAN2ON);}
    else {PORTB&=(unsigned)~(1<<PINB_FAN2ON);}
    OCR1B=dC2;
    
    uint8_t sc= senseChanged_I;
    if(tDC1)
    {
        if(in_Sense_Speed_I^flag_in_Sense_Speed_I) {sc++;}
        if(!systemTimer)
        {
            fanBlocked_I=(sc<3);
            sc=0;
        }
    }
    else if(!fanBlocked_I) {sc=128;}
    senseChanged_I=sc;
    flag_in_Sense_Speed_I=in_Sense_Speed_I;

    sc= senseChanged_II;
    if(tDC2)
    {
        if(in_Sense_Speed_II^flag_in_Sense_Speed_II) {sc++;}
        if(!systemTimer)
        {
            fanBlocked_II=(sc<3);
            sc=0;
        }
    }
    else if(!fanBlocked_II) {sc=128;}
    senseChanged_II=sc;
    flag_in_Sense_Speed_II=in_Sense_Speed_II;

    if(!systemTimer) {systemTimer=10;}
}

int main(void)
{
	PORTD=(1<<PIND_MASTER); //switch on pull up
	asm("nop");
	in_master=!((PIND>>(PIND_MASTER))&1);
	
	PORTD=(1<<PIND_MASTER)|(1<<PIND_OPEN)|(1<<PIND_SENSE_SPEED_I)|(1<<PIND_SENSE_SPEED_II); //switch on pull up  (v_in has external pull down)
	if(in_master)
	{
    	PORTB=(1<<PINB_NC_1)|(1<<PINB_NC_2)|(1<<PINB_NC_3)|(1<<PINB_NC_4); //switch on pull up
    	DDRB=(1<<PINB_SUMMER);
    	DDRD=(1<<PIND_V_OUT);
        systemTimer=100;
	}
	else
	{
    	DDRB=(1<<PINB_PWM2)|(1<<PINB_PWM1)|(1<<PINB_FAN2ON)|(1<<PINB_FAN1ON)|(1<<PINB_ERR);
    	//pwm mode
    	TCCR1A=(1<<COM1A1)|(1<<COM1B1)|(1<<WGM10);  //phase correct PWM 8-bit mode on OC1A and OC1B
    	OCR1A=0x00; //initial PWM pulse width = 0
    	OCR1B=0x00; //initial PWM pulse width = 0
    	TCCR1B=(1<<CS10);  //clock source = clkI/O/1 (no prescale), start PWM
        connected=connectTimeout;
        systemTimer=10;
	}

	// 1ms Timer
	TCCR0A=(1<<WGM01); //CTC Mode
	TCCR0B|=((1<<CS02)|(1<<CS00)); // prescale 1024
	OCR0A=11;
	TIMSK|=(1<<OCIE0A); //Compare Interrupt on counter 0A
	
	sei();
    
    while(1) 
    {	
        if(flag_ms_1)
        {
            flag_ms_1=false;
            in_Sense_Speed_II=!((PIND>>PIND_SENSE_SPEED_II)&1);
            uint8_t iSSI=!((PIND>>PIND_SENSE_SPEED_I)&1);
            
            if(in_master)
            {
                //debounce speed I to void wrong detection of summer mode
                uint8_t iSSIs=in_Sense_Speed_I_shift;
                (iSSIs<<=1)|=iSSI;
                if(!iSSIs) {in_Sense_Speed_I=false;}
                else if(iSSIs==0xFF) {in_Sense_Speed_I=true;}
                in_Sense_Speed_I_shift=iSSIs;
                masterMain();
            }
            else                
            {
                in_Sense_Speed_I=iSSI;
                in_V_IN=(PIND>>PIND_V_IN)&1;
                slaveMain();
            }
        }        
    }    
}

ISR(TIMER0_COMPA_vect) //1 ms interrupt
{
    flag_ms_1=true;
    if(ms_100) {ms_100--;} else
    {
        ms_100=100;
        if(systemTimer) {systemTimer--;}
        if(modeSummerTestGate) {modeSummerTestGate--;}
    }
    chkPeriode++;
    if(in_master&&(chkPeriode>=250)) {chkPeriode=0;}
}
